@@ -1,28 +1,144 @@
 import datetime
 import functools
-import inspect
 import sys
 import time
+import uuid
 import calendar
 import unittest
 import platform
 import warnings
+import types
+import numbers
+import inspect
 
 from dateutil import parser
 from dateutil.tz import tzlocal
 
+
+try:
+    from maya import MayaDT
+except ImportError:
+    MayaDT = None
+
+_TIME_NS_PRESENT = hasattr(time, 'time_ns')
+
 real_time = time.time
 real_localtime = time.localtime
 real_gmtime = time.gmtime
 real_strftime = time.strftime
+real_timegm = calendar.timegm
 real_date = datetime.date
 real_datetime = datetime.datetime
+real_date_objects = [real_time, real_localtime, real_gmtime, real_strftime, real_timegm, real_date, real_datetime]
+
+if _TIME_NS_PRESENT:
+    real_time_ns = time.time_ns
+    real_date_objects.append(real_time_ns)
+
+_real_time_object_ids = {id(obj) for obj in real_date_objects}
+
+# time.clock is deprecated and was removed in Python 3.8
+real_clock = getattr(time, 'clock', None)
+
+freeze_factories = []
+tz_offsets = []
+ignore_lists = []
+tick_flags = []
+
+# Python3 doesn't have basestring, but it does have str.
+try:
+    # noinspection PyUnresolvedReferences
+    _string_type = basestring
+except NameError:
+    _string_type = str
+
+try:
+    # noinspection PyUnresolvedReferences
+    real_uuid_generate_time = uuid._uuid_generate_time
+    uuid_generate_time_attr = '_uuid_generate_time'
+except AttributeError:
+    # noinspection PyUnresolvedReferences
+    uuid._load_system_functions()
+    # noinspection PyUnresolvedReferences
+    real_uuid_generate_time = uuid._generate_time_safe
+    uuid_generate_time_attr = '_generate_time_safe'
+except ImportError:
+    real_uuid_generate_time = None
+    uuid_generate_time_attr = None
+
+try:
+    # noinspection PyUnresolvedReferences
+    real_uuid_create = uuid._UuidCreate
+except (AttributeError, ImportError):
+    real_uuid_create = None
 
 try:
     import copy_reg as copyreg
 except ImportError:
     import copyreg
 
+try:
+    iscoroutinefunction = inspect.iscoroutinefunction
+    if sys.version_info < (3, 5):
+        from freezegun._async_coroutine import wrap_coroutine
+    else:
+        from freezegun._async import wrap_coroutine
+except AttributeError:
+    iscoroutinefunction = lambda x: False
+
+    def wrap_coroutine(*args):
+        raise NotImplementedError()
+
+
+# keep a cache of module attributes otherwise freezegun will need to analyze too many modules all the time
+_GLOBAL_MODULES_CACHE = {}
+
+
+def _get_module_attributes(module):
+    result = []
+    try:
+        module_attributes = dir(module)
+    except (ImportError, TypeError):
+        return result
+    for attribute_name in module_attributes:
+        try:
+            attribute_value = getattr(module, attribute_name)
+        except (ImportError, AttributeError, TypeError):
+            # For certain libraries, this can result in ImportError(_winreg) or AttributeError (celery)
+            continue
+        else:
+            result.append((attribute_name, attribute_value))
+    return result
+
+
+def _setup_module_cache(module):
+    date_attrs = []
+    all_module_attributes = _get_module_attributes(module)
+    for attribute_name, attribute_value in all_module_attributes:
+        if id(attribute_value) in _real_time_object_ids:
+            date_attrs.append((attribute_name, attribute_value))
+    _GLOBAL_MODULES_CACHE[module.__name__] = (_get_module_attributes_hash(module), date_attrs)
+
+
+def _get_module_attributes_hash(module):
+    try:
+        module_dir = dir(module)
+    except (ImportError, TypeError):
+        module_dir = []
+    return '{}-{}'.format(id(module), hash(frozenset(module_dir)))
+
+
+def _get_cached_module_attributes(module):
+    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE.get(module.__name__, ('0', []))
+    if _get_module_attributes_hash(module) == module_hash:
+        return cached_attrs
+
+    # cache miss: update the cache and return the refreshed value
+    _setup_module_cache(module)
+    # return the newly cached value
+    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE[module.__name__]
+    return cached_attrs
+
 
 # Stolen from six
 def with_metaclass(meta, *bases):
@@ -35,56 +151,111 @@ _is_cpython = (
 )
 
 
-class FakeTime(object):
-
-    def __init__(self, time_to_freeze, previous_time_function):
-        self.time_to_freeze = time_to_freeze
-        self.previous_time_function = previous_time_function
-
-    def __call__(self):
-        current_time = self.time_to_freeze()
-        return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
+call_stack_inspection_limit = 5
 
 
-class FakeLocalTime(object):
-    def __init__(self, time_to_freeze, previous_localtime_function=None):
-        self.time_to_freeze = time_to_freeze
-        self.previous_localtime_function = previous_localtime_function
+def _should_use_real_time():
+    if not call_stack_inspection_limit:
+        return False
 
-    def __call__(self, t=None):
-        if t is not None:
-            return real_localtime(t)
-        shifted_time = self.time_to_freeze() - datetime.timedelta(seconds=time.timezone)
-        return shifted_time.timetuple()
+    if not ignore_lists[-1]:
+        return False
+
+    frame = inspect.currentframe().f_back.f_back
+
+    for _ in range(call_stack_inspection_limit):
+        module_name = frame.f_globals.get('__name__')
+        if module_name and module_name.startswith(ignore_lists[-1]):
+            return True
+
+        frame = frame.f_back
+        if frame is None:
+            break
+
+    return False
 
 
-class FakeGMTTime(object):
-    def __init__(self, time_to_freeze, previous_gmtime_function):
-        self.time_to_freeze = time_to_freeze
-        self.previous_gmtime_function = previous_gmtime_function
-
-    def __call__(self, t=None):
-        if t is not None:
-            return real_gmtime(t)
-        return self.time_to_freeze().timetuple()
+def get_current_time():
+    return freeze_factories[-1]()
 
 
-class FakeStrfTime(object):
-    def __init__(self, time_to_freeze, previous_strftime_function):
-        self.time_to_freeze = time_to_freeze
-        self.previous_strftime_function = previous_strftime_function
+def fake_time():
+    if _should_use_real_time():
+        return real_time()
+    current_time = get_current_time()
+    return real_timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0
 
-    def __call__(self, format, time_to_format=None):
-        if time_to_format is None:
-            time_to_format = FakeLocalTime(self.time_to_freeze)()
+if _TIME_NS_PRESENT:
+    def fake_time_ns():
+        if _should_use_real_time():
+            return real_time_ns()
+        return int(int(fake_time()) * 1e9)
+
+
+def fake_localtime(t=None):
+    if t is not None:
+        return real_localtime(t)
+    if _should_use_real_time():
+        return real_localtime()
+    shifted_time = get_current_time() - datetime.timedelta(seconds=time.timezone)
+    return shifted_time.timetuple()
+
+
+def fake_gmtime(t=None):
+    if t is not None:
+        return real_gmtime(t)
+    if _should_use_real_time():
+        return real_gmtime()
+    return get_current_time().timetuple()
+
+
+def fake_strftime(format, time_to_format=None):
+    if time_to_format is None:
+        if not _should_use_real_time():
+            time_to_format = fake_localtime()
+
+    if time_to_format is None:
+        return real_strftime(format)
+    else:
         return real_strftime(format, time_to_format)
 
 
+def fake_timegm(struct_time):
+    if _should_use_real_time():
+        return real_timegm(struct_time)
+    else:
+        return real_timegm(get_current_time().timetuple())
+
+
+if real_clock is not None:
+    def fake_clock():
+        if _should_use_real_time():
+            return real_clock()
+
+        if len(freeze_factories) == 1:
+            return 0.0 if not tick_flags[-1] else real_clock()
+
+        first_frozen_time = freeze_factories[0]()
+        last_frozen_time = get_current_time()
+
+        timedelta = (last_frozen_time - first_frozen_time)
+        total_seconds = timedelta.total_seconds()
+
+        if tick_flags[-1]:
+            total_seconds += real_clock()
+
+        return total_seconds
+
+
 class FakeDateMeta(type):
     @classmethod
     def __instancecheck__(self, obj):
         return isinstance(obj, real_date)
 
+    @classmethod
+    def __subclasscheck__(cls, subclass):
+        return issubclass(subclass, real_date)
+
 
 def datetime_to_fakedatetime(datetime):
     return FakeDatetime(datetime.year,
@@ -104,9 +275,6 @@ def date_to_fakedate(date):
 
 
 class FakeDate(with_metaclass(FakeDateMeta, real_date)):
-    dates_to_freeze = []
-    tz_offsets = []
-
     def __new__(cls, *args, **kwargs):
         return real_date.__new__(cls, *args, **kwargs)
 
@@ -127,16 +295,16 @@ class FakeDate(with_metaclass(FakeDateMeta, real_date)):
 
     @classmethod
     def today(cls):
-        result = cls._date_to_freeze() + datetime.timedelta(hours=cls._tz_offset())
+        result = cls._date_to_freeze() + cls._tz_offset()
         return date_to_fakedate(result)
 
-    @classmethod
-    def _date_to_freeze(cls):
-        return cls.dates_to_freeze[-1]()
+    @staticmethod
+    def _date_to_freeze():
+        return get_current_time()
 
     @classmethod
     def _tz_offset(cls):
-        return cls.tz_offsets[-1]
+        return tz_offsets[-1]
 
 FakeDate.min = date_to_fakedate(real_date.min)
 FakeDate.max = date_to_fakedate(real_date.max)
@@ -147,11 +315,12 @@ class FakeDatetimeMeta(FakeDateMeta):
     def __instancecheck__(self, obj):
         return isinstance(obj, real_datetime)
 
+    @classmethod
+    def __subclasscheck__(cls, subclass):
+        return issubclass(subclass, real_datetime)
+
 
 class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
-    times_to_freeze = []
-    tz_offsets = []
-
     def __new__(cls, *args, **kwargs):
         return real_datetime.__new__(cls, *args, **kwargs)
 
@@ -177,31 +346,42 @@ class FakeDatetime(with_metaclass(FakeDatetimeMeta, real_datetime, FakeDate)):
 
     @classmethod
     def now(cls, tz=None):
+        now = cls._time_to_freeze() or real_datetime.now()
         if tz:
-            result = tz.fromutc(cls._time_to_freeze().replace(tzinfo=tz)) + datetime.timedelta(hours=cls._tz_offset())
+            result = tz.fromutc(now.replace(tzinfo=tz)) + cls._tz_offset()
         else:
-            result = cls._time_to_freeze() + datetime.timedelta(hours=cls._tz_offset())
+            result = now + cls._tz_offset()
         return datetime_to_fakedatetime(result)
 
     def date(self):
         return date_to_fakedate(self)
 
+    @property
+    def nanosecond(self):
+        try:
+            # noinspection PyUnresolvedReferences
+            return real_datetime.nanosecond
+        except AttributeError:
+            return 0
+
     @classmethod
     def today(cls):
         return cls.now(tz=None)
 
     @classmethod
     def utcnow(cls):
-        result = cls._time_to_freeze()
+        result = cls._time_to_freeze() or real_datetime.utcnow()
         return datetime_to_fakedatetime(result)
 
-    @classmethod
-    def _time_to_freeze(cls):
-        return cls.times_to_freeze[-1]()
+    @staticmethod
+    def _time_to_freeze():
+        if freeze_factories:
+            return get_current_time()
 
     @classmethod
     def _tz_offset(cls):
-        return cls.tz_offsets[-1]
+        return tz_offsets[-1]
+
 
 FakeDatetime.min = datetime_to_fakedatetime(real_datetime.min)
 FakeDatetime.max = datetime_to_fakedatetime(real_datetime.max)
@@ -251,12 +431,21 @@ def _parse_time_to_freeze(time_to_freeze_str):
         time_to_freeze = time_to_freeze_str
     elif isinstance(time_to_freeze_str, datetime.date):
         time_to_freeze = datetime.datetime.combine(time_to_freeze_str, datetime.time())
+    elif isinstance(time_to_freeze_str, datetime.timedelta):
+        time_to_freeze = datetime.datetime.utcnow() + time_to_freeze_str
     else:
         time_to_freeze = parser.parse(time_to_freeze_str)
 
     return convert_to_timezone_naive(time_to_freeze)
 
 
+def _parse_tz_offset(tz_offset):
+    if isinstance(tz_offset, datetime.timedelta):
+        return tz_offset
+    else:
+        return datetime.timedelta(hours=tz_offset)
+
+
 class TickingDateTimeFactory(object):
 
     def __init__(self, time_to_freeze, start):
@@ -276,8 +465,38 @@ class FrozenDateTimeFactory(object):
         return self.time_to_freeze
 
     def tick(self, delta=datetime.timedelta(seconds=1)):
+        if isinstance(delta, numbers.Real):
+            # noinspection PyTypeChecker
+            self.time_to_freeze += datetime.timedelta(seconds=delta)
+        else:
+            self.time_to_freeze += delta
+
+    def move_to(self, target_datetime):
+        """Moves frozen date to the given ``target_datetime``"""
+        target_datetime = _parse_time_to_freeze(target_datetime)
+        delta = target_datetime - self.time_to_freeze
+        self.tick(delta=delta)
+
+
+class StepTickTimeFactory(object):
+
+    def __init__(self, time_to_freeze, step_width):
+        self.time_to_freeze = time_to_freeze
+        self.step_width = step_width
+
+    def __call__(self):
+        return_time = self.time_to_freeze
+        self.tick()
+        return return_time
+
+    def tick(self, delta=None):
+        if not delta:
+            delta = datetime.timedelta(seconds=self.step_width)
         self.time_to_freeze += delta
 
+    def update_step_width(self, step_width):
+        self.step_width = step_width
+
     def move_to(self, target_datetime):
         """Moves frozen date to the given ``target_datetime``"""
         target_datetime = _parse_time_to_freeze(target_datetime)
@@ -287,18 +506,22 @@ class FrozenDateTimeFactory(object):
 
 class _freeze_time(object):
 
-    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick):
 
+    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, auto_tick_seconds):
         self.time_to_freeze = _parse_time_to_freeze(time_to_freeze_str)
-        self.tz_offset = tz_offset
+        self.tz_offset = _parse_tz_offset(tz_offset)
         self.ignore = tuple(ignore)
         self.tick = tick
+        self.auto_tick_seconds = auto_tick_seconds
         self.undo_changes = []
         self.modules_at_start = set()
+        self.as_arg = as_arg
 
     def __call__(self, func):
         if inspect.isclass(func):
             return self.decorate_class(func)
+        elif iscoroutinefunction(func):
+            return self.decorate_coroutine(func)
         return self.decorate_callable(func)
 
     def decorate_class(self, klass):
@@ -310,12 +533,14 @@ class _freeze_time(object):
             orig_setUpClass = getattr(klass, 'setUpClass', None)
             orig_tearDownClass = getattr(klass, 'tearDownClass', None)
 
+            # noinspection PyDecorator
             @classmethod
             def setUpClass(cls):
                 self.start()
                 if orig_setUpClass is not None:
                     orig_setUpClass()
 
+            # noinspection PyDecorator
             @classmethod
             def tearDownClass(cls):
                 if orig_tearDownClass is not None:
@@ -355,39 +580,64 @@ class _freeze_time(object):
         self.stop()
 
     def start(self):
-        if self.tick:
-            time_to_freeze = TickingDateTimeFactory(self.time_to_freeze, real_datetime.now())
+
+        if self.auto_tick_seconds:
+            freeze_factory = StepTickTimeFactory(self.time_to_freeze, self.auto_tick_seconds)
+        elif self.tick:
+            freeze_factory = TickingDateTimeFactory(self.time_to_freeze, real_datetime.now())
         else:
-            time_to_freeze = FrozenDateTimeFactory(self.time_to_freeze)
+            freeze_factory = FrozenDateTimeFactory(self.time_to_freeze)
+
+        is_already_started = len(freeze_factories) > 0
+        freeze_factories.append(freeze_factory)
+        tz_offsets.append(self.tz_offset)
+        ignore_lists.append(self.ignore)
+        tick_flags.append(self.tick)
+
+        if is_already_started:
+            return freeze_factory
 
         # Change the modules
+        calendar.timegm = fake_timegm
+
         datetime.datetime = FakeDatetime
         datetime.date = FakeDate
-        fake_time = FakeTime(time_to_freeze, time.time)
-        fake_localtime = FakeLocalTime(time_to_freeze, time.localtime)
-        fake_gmtime = FakeGMTTime(time_to_freeze, time.gmtime)
-        fake_strftime = FakeStrfTime(time_to_freeze, time.strftime)
+
         time.time = fake_time
         time.localtime = fake_localtime
         time.gmtime = fake_gmtime
         time.strftime = fake_strftime
+        if uuid_generate_time_attr:
+            setattr(uuid, uuid_generate_time_attr, None)
+        uuid._UuidCreate = None
+        uuid._last_timestamp = None
 
         copyreg.dispatch_table[real_datetime] = pickle_fake_datetime
         copyreg.dispatch_table[real_date] = pickle_fake_date
 
         # Change any place where the module had already been imported
         to_patch = [
-            ('real_date', real_date, 'FakeDate', FakeDate),
-            ('real_datetime', real_datetime, 'FakeDatetime', FakeDatetime),
-            ('real_gmtime', real_gmtime, 'FakeGMTTime', fake_gmtime),
-            ('real_localtime', real_localtime, 'FakeLocalTime', fake_localtime),
-            ('real_strftime', real_strftime, 'FakeStrfTime', fake_strftime),
-            ('real_time', real_time, 'FakeTime', fake_time),
+            ('real_date', real_date, FakeDate),
+            ('real_datetime', real_datetime, FakeDatetime),
+            ('real_gmtime', real_gmtime, fake_gmtime),
+            ('real_localtime', real_localtime, fake_localtime),
+            ('real_strftime', real_strftime, fake_strftime),
+            ('real_time', real_time, fake_time),
+            ('real_timegm', real_timegm, fake_timegm),
         ]
-        real_names = tuple(real_name for real_name, real, fake_name, fake in to_patch)
-        self.fake_names = tuple(fake_name for real_name, real, fake_name, fake in to_patch)
-        self.reals = dict((id(fake), real) for real_name, real, fake_name, fake in to_patch)
-        fakes = dict((id(real), fake) for real_name, real, fake_name, fake in to_patch)
+
+        if _TIME_NS_PRESENT:
+            time.time_ns = fake_time_ns
+            to_patch.append(('real_time_ns', real_time_ns, fake_time_ns))
+
+        if real_clock is not None:
+            # time.clock is deprecated and was removed in Python 3.8
+            time.clock = fake_clock
+            to_patch.append(('real_clock', real_clock, fake_clock))
+
+        self.fake_names = tuple(fake.__name__ for real_name, real, fake in to_patch)
+        self.reals = {id(fake): real for real_name, real, fake in to_patch}
+        fakes = {id(real): fake for real_name, real, fake in to_patch}
         add_change = self.undo_changes.append
 
         # Save the current loaded modules
@@ -397,40 +647,30 @@ class _freeze_time(object):
             warnings.filterwarnings('ignore')
 
             for mod_name, module in list(sys.modules.items()):
-                if mod_name is None or module is None:
+                if mod_name is None or module is None or mod_name == __name__:
                     continue
-                elif mod_name.startswith(self.ignore):
+                elif mod_name.startswith(self.ignore) or mod_name.endswith('.six.moves'):
                     continue
                 elif (not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time')):
                     continue
-                for module_attribute in dir(module):
-                    if module_attribute in real_names:
-                        continue
-                    try:
-                        attribute_value = getattr(module, module_attribute)
-                    except (ImportError, AttributeError, TypeError):
-                        # For certain libraries, this can result in ImportError(_winreg) or AttributeError (celery)
-                        continue
+
+                module_attrs = _get_cached_module_attributes(module)
+                for attribute_name, attribute_value in module_attrs:
                     fake = fakes.get(id(attribute_value))
                     if fake:
-                        setattr(module, module_attribute, fake)
-                        add_change((module, module_attribute, attribute_value))
+                        setattr(module, attribute_name, fake)
+                        add_change((module, attribute_name, attribute_value))
 
-        datetime.datetime.times_to_freeze.append(time_to_freeze)
-        datetime.datetime.tz_offsets.append(self.tz_offset)
-
-        datetime.date.dates_to_freeze.append(time_to_freeze)
-        datetime.date.tz_offsets.append(self.tz_offset)
-
-        return time_to_freeze
+        return freeze_factory
 
     def stop(self):
-        datetime.datetime.times_to_freeze.pop()
-        datetime.datetime.tz_offsets.pop()
-        datetime.date.dates_to_freeze.pop()
-        datetime.date.tz_offsets.pop()
+        freeze_factories.pop()
+        ignore_lists.pop()
+        tick_flags.pop()
+        tz_offsets.pop()
 
-        if not datetime.datetime.times_to_freeze:
+        if not freeze_factories:
+            calendar.timegm = real_timegm
             datetime.datetime = real_datetime
             datetime.date = real_date
             copyreg.dispatch_table.pop(real_datetime)
@@ -448,7 +688,7 @@ class _freeze_time(object):
                     module = sys.modules.get(mod_name, None)
                     if mod_name is None or module is None:
                         continue
-                    elif mod_name.startswith(self.ignore):
+                    elif mod_name.startswith(self.ignore) or mod_name.endswith('.six.moves'):
                         continue
                     elif (not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time')):
                         continue
@@ -466,15 +706,30 @@ class _freeze_time(object):
                         if real:
                             setattr(module, module_attribute, real)
 
-        time.time = time.time.previous_time_function
-        time.gmtime = time.gmtime.previous_gmtime_function
-        time.localtime = time.localtime.previous_localtime_function
-        time.strftime = time.strftime.previous_strftime_function
+            time.time = real_time
+            time.gmtime = real_gmtime
+            time.localtime = real_localtime
+            time.strftime = real_strftime
+            time.clock = real_clock
+
+            if _TIME_NS_PRESENT:
+                time.time_ns = real_time_ns
+
+            if uuid_generate_time_attr:
+                setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
+            uuid._UuidCreate = real_uuid_create
+            uuid._last_timestamp = None
+
+    def decorate_coroutine(self, coroutine):
+        return wrap_coroutine(self, coroutine)
 
     def decorate_callable(self, func):
         def wrapper(*args, **kwargs):
-            with self:
-                result = func(*args, **kwargs)
+            with self as time_factory:
+                if self.as_arg:
+                    result = func(time_factory, *args, **kwargs)
+                else:
+                    result = func(*args, **kwargs)
             return result
         functools.update_wrapper(wrapper, func)
 
@@ -485,30 +740,49 @@ class _freeze_time(object):
         return wrapper
 
 
-def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False):
-    # Python3 doesn't have basestring, but it does have str.
-    try:
-        string_type = basestring
-    except NameError:
-        string_type = str
+def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, auto_tick_seconds=0):
+    acceptable_times = (type(None), _string_type, datetime.date, datetime.timedelta,
+             types.FunctionType, types.GeneratorType)
 
-    if not isinstance(time_to_freeze, (type(None), string_type, datetime.date)):
-        raise TypeError(('freeze_time() expected None, a string, date instance, or '
-                         'datetime instance, but got type {0}.').format(type(time_to_freeze)))
+    if MayaDT is not None:
+        acceptable_times += MayaDT,
+
+    if not isinstance(time_to_freeze, acceptable_times):
+        raise TypeError(('freeze_time() expected None, a string, date instance, datetime '
+                         'instance, MayaDT, timedelta instance, function or a generator, but got '
+                         'type {}.').format(type(time_to_freeze)))
     if tick and not _is_cpython:
         raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')
 
+    if isinstance(time_to_freeze, types.FunctionType):
+        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, auto_tick_seconds)
+
+    if isinstance(time_to_freeze, types.GeneratorType):
+        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, auto_tick_seconds)
+
+    if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
+        return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
+                           tick, as_arg)
+
     if ignore is None:
         ignore = []
+    ignore = ignore[:]
+    ignore.append('nose.plugins')
     ignore.append('six.moves')
     ignore.append('django.utils.six.moves')
+    ignore.append('google.gax')
     ignore.append('threading')
     ignore.append('Queue')
-    return _freeze_time(time_to_freeze, tz_offset, ignore, tick)
+    ignore.append('selenium')
+    ignore.append('_pytest.terminal.')
+    ignore.append('_pytest.runner.')
+
+    return _freeze_time(time_to_freeze, tz_offset, ignore, tick, as_arg, auto_tick_seconds)
 
 
 # Setup adapters for sqlite
 try:
+    # noinspection PyUnresolvedReferences
     import sqlite3
 except ImportError:
     # Some systems have trouble with this
